#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdarg.h>
#include <pthread.h>

#include <mach/mach.h>
#include <mach/task.h>
#include <mach/mach_error.h>
#include <mach/mach_traps.h>

#include "gadgets.h"

#include "remote_call.h"
#include "remote_memory.h"

// this code is based on what I wrote for triple_fetch, but changed to only require a thread port rather
// than a task port

void test_gadgets() {
  char* candidates[] = {"\xa0\x47", NULL}; // thumb: blx r4
  find_gadget_candidate(candidates, 2);
}

void test_thread_control(mach_port_t thread_port) {
  _STRUCT_ARM_THREAD_STATE thread_state = {0};
  //arm_thread_state32_t thread_state = {0};
  mach_msg_type_number_t thread_stateCnt = sizeof(thread_state)/4;
  kern_return_t err;
  err = thread_suspend(thread_port);
  if (err != KERN_SUCCESS) {
    printf("thread_suspend failed: %x (%s)\n", err, mach_error_string(err));
  }

  err = thread_get_state(thread_port, ARM_THREAD_STATE, (thread_state_t)&thread_state, &thread_stateCnt);
  if (err != KERN_SUCCESS) {
    printf("thread_get_state failed: %x (%s)\n", err, mach_error_string(err));
  }
  
  printf("thread pc: 0x%08x\n", thread_state.__pc);
  
  thread_state.__pc = 0x41424344;
  
  err = thread_set_state(thread_port, ARM_THREAD_STATE, (thread_state_t)&thread_state, thread_stateCnt);
  if (err != KERN_SUCCESS) {
    printf("thread_set_state failed: %x (%s)\n", err, mach_error_string(err));
  }
  
  printf("set state\n");
  
  err = thread_resume(thread_port);
  if (err != KERN_SUCCESS) {
    printf("thread_resume failed: %x (%s)\n", err, mach_error_string(err));
  }
  
  
  printf("resumed\n");
  
  sleep(100);
}

uint32_t blx_r4_addr = 0;

uint32_t
find_blx_r4_gadget() {
  if (blx_r4_addr != 0) {
    return blx_r4_addr;
  }
  
  char* candidates[] = {"\xa0\x47", NULL}; // thumb: blx r4
  blx_r4_addr = find_gadget_candidate(candidates, 2);
  return blx_r4_addr;
}



// prepare the thread we'll use for doing arbitrary remote calls
//mach_port_t target_thread_port = MACH_PORT_NULL;
//uint64_t looper_lr = 0;
#if 0
void prepare_thread_call(mach_port_t thread_port) {
  //target_thread_port = thread_port;
  
  kern_return_t err;
  
  // suspend the target thread we'll hijack:
  err = thread_suspend(thread_port);
  if (err != KERN_SUCCESS) {
    printf("failed to suspend the thread we're trying to hijack: %s\n", mach_error_string(err));
    return;
  }
  
  // save its suspended state so we can restore it:
  _STRUCT_ARM_THREAD_STATE64 thread_state = {0};
  mach_msg_type_number_t thread_stateCnt = sizeof(thread_state)/4;
  printf("saved thread state count before: %d\n", thread_stateCnt);
  err = thread_get_state(thread_port, ARM_THREAD_STATE64, (thread_state_t)&thread_state, &thread_stateCnt);
  if (err != KERN_SUCCESS){
    printf("error getting thread state to save: %s\n", mach_error_string(err));
    return;
  }
  
  // start the thread spinning, we'll just leave it in this state
  looper_lr = find_blr_x19_gadget();
  printf("looper_lr: %llx\n", looper_lr);
  thread_state.__x[19] = looper_lr;
  thread_state.__lr = looper_lr;
  
  // set the thread state:
  err = thread_set_state(thread_port, ARM_THREAD_STATE64, (thread_state_t)&thread_state, thread_stateCnt);
  if (err != KERN_SUCCESS){
    printf("error setting new thread state for hijacked thread: %s\n", mach_error_string(err));
    return;
  }
  
  printf("target thread is suspended\n");
}
#endif

void end_thread(mach_port_t thread_port) {
#if 0
  kern_return_t err;
  
  // save its suspended state so we can restore it:
  _STRUCT_ARM_THREAD_STATE thread_state = {0};
  mach_msg_type_number_t thread_stateCnt = sizeof(thread_state)/4;
  printf("saved thread state count before: %d\n", thread_stateCnt);
  err = thread_get_state(thread_port, ARM_THREAD_STATE, (thread_state_t)&thread_state, &thread_stateCnt);
  if (err != KERN_SUCCESS){
    printf("error getting thread state to save: %s\n", mach_error_string(err));
    return;
  }
  
  thread_state.__pc = 0xffffffffffff;
  
  // set the thread state:
  err = thread_set_state(thread_port, ARM_THREAD_STATE64, (thread_state_t)&thread_state, thread_stateCnt);
  if (err != KERN_SUCCESS){
    printf("error setting new thread state for hijacked thread: %s\n", mach_error_string(err));
    return;
  }
  
  printf("set thread with pc: %llx\n", thread_state.__pc);
  
  err = thread_resume(thread_port);
  if (err != KERN_SUCCESS) {
    printf("end_thread failed to resume: %x %s\n", err, mach_error_string(err));
  }
#endif
  thread_call_remote(thread_port, (void*)0x123456789, 0);
}



// no support for non-register args
#define MAX_REMOTE_ARGS 8
// thread should be suspended already; will return suspended
uint32_t thread_call_remote(mach_port_t thread_port, void* fptr, int n_params, ...)
{
  if (n_params > MAX_REMOTE_ARGS || n_params < 0){
    printf("unsupported number of arguments to remote function (%d)\n", n_params);
    return 0;
  }
  
  kern_return_t err;
 
  
  //#if 0
  // suspend the target thread we'll hijack:
  err = thread_suspend(thread_port);
  if (err != KERN_SUCCESS) {
    printf("failed to suspend the thread we're trying to hijack: %s\n", mach_error_string(err));
    return 0;
  }
  //#endif
  
  // save its suspended state so we can restore it:
  _STRUCT_ARM_THREAD_STATE saved_thread_state = {0};
  mach_msg_type_number_t saved_thread_stateCnt = sizeof(saved_thread_state)/4;
  // printf("saved thread state count before: %d\n", saved_thread_stateCnt);
  err = thread_get_state(thread_port, ARM_THREAD_STATE, (thread_state_t)&saved_thread_state, &saved_thread_stateCnt);
  if (err != KERN_SUCCESS){
    printf("error getting thread state to save: %s\n", mach_error_string(err));
    return 0;
  }
  //printf("saved thread state count after: %d\n", saved_thread_stateCnt);
  
  // build the state we need for the arbitrary call:
  _STRUCT_ARM_THREAD_STATE fcall_thread_state = {0};
  mach_msg_type_number_t fcall_thread_stateCnt = sizeof(fcall_thread_state)/4;
  memcpy(&fcall_thread_state, &saved_thread_state, sizeof(fcall_thread_state));
  
  // jump down the stack a bit
  //fcall_thread_state.__sp -= 0x100;
  
  
  // make sure we can determine when the function call is done
  // the gadget is thumb
  fcall_thread_state.__r[4] = find_blx_r4_gadget() | 1;
  fcall_thread_state.__lr = find_blx_r4_gadget() | 1;
  
  // set the pc
  fcall_thread_state.__pc = (uint32_t)fptr;
  
  //fcall_thread_state.__cpsr |= 0x20; // make sure it actually goes in to thumb mode!
  
  if ((uint32_t)fptr & 1) {
    fcall_thread_state.__cpsr |= 0x20;
  }
  
  // load the arguments
  va_list ap;
  va_start(ap, n_params);
  
  arg_desc* args[MAX_REMOTE_ARGS] = {0};
  
  uint64_t remote_buffers[MAX_REMOTE_ARGS] = {0};
  
  for (int i = 0; i < n_params; i++){
    arg_desc* arg = va_arg(ap, arg_desc*);
    
    args[i] = arg;
    
    switch(arg->type){
      case ARG_LITERAL:
      {
        //printf("setting arg %d to literal %llx\n", i, arg->value);
        fcall_thread_state.__r[i] = arg->value;
        break;
      }
#if 0
      case ARG_BUFFER:
      case ARG_BUFFER_PERSISTENT:
      {
        uint64_t remote_buffer = alloc_and_fill_remote_buffer(task_port, arg->value, arg->length);
        remote_buffers[i] = remote_buffer;
        thread_state.__x[i] = remote_buffer;
        break;
      }
        
      case ARG_OUT_BUFFER:
      {
        uint64_t remote_buffer = remote_alloc(task_port, arg->length);
        printf("allocated a remote out buffer: %llx\n", remote_buffer);
        remote_buffers[i] = remote_buffer;
        thread_state.__x[i] = remote_buffer;
        break;
      }
#endif
      default:
      {
        printf("invalid argument type!\n");
      }
    }
  }
  
  va_end(ap);
//#if 0
  printf("fcall thread state:\n");
  printf("pc: 0x%08x\n", fcall_thread_state.__pc);
  printf("sp: 0x%08x\n", fcall_thread_state.__sp);
  printf("lr: 0x%08x\n", fcall_thread_state.__lr);
  printf("cpsr: 0x%08x\n", fcall_thread_state.__cpsr);
  for (int i = 0; i < 13; i++) {
    printf("x%d: 0x%08x\n", i, fcall_thread_state.__r[i]);
  }
//#endif
  
  // set the thread state:
  err = thread_set_state(thread_port, ARM_THREAD_STATE, (thread_state_t)&fcall_thread_state, fcall_thread_stateCnt);
  if (err != KERN_SUCCESS){
    printf("error setting new thread state for hijacked thread: %s\n", mach_error_string(err));
    return 0;
  }
  
  // let the thread continue running with the new state:
  err = thread_resume(thread_port);
  if (err != KERN_SUCCESS) {
    printf("error resuming hijacked thread: %s\n", mach_error_string(err));
    return 0;
  }
  printf("resumed thread\n");
  
  // monitor for the function call ending and the thread hitting the infinite loop:
  // we're reusing fcall state so we can also get the return value via x0
  while(1){
    usleep(100*1000);
    thread_suspend(thread_port);
    err = thread_get_state(thread_port, ARM_THREAD_STATE, (thread_state_t)&fcall_thread_state, &fcall_thread_stateCnt);
    if (err != KERN_SUCCESS){
      printf("error getting thread state: %s\n", mach_error_string(err));
      return 0;
    }
    
    thread_resume(thread_port);
    
    if (fcall_thread_state.__pc == find_blx_r4_gadget()){
      // thread has returned from the target function
      printf("hit looper!\n");
      break;
    }
    printf("got bad pc: 0x%08x\n", fcall_thread_state.__pc);
  }
  
  uint32_t ret_val = fcall_thread_state.__r[0];
  printf("ret val: %llx\n", ret_val);
  
  
  // thread is already suspended
  
  //thread_suspend(thread_port);
  
  //printf("suspended the thread...\n");
  //sleep(100);
  //thread_resume(thread_port);
#if 0
  // suspend the target thread:
  err = thread_suspend(thread_port);
  if (err != KERN_SUCCESS) {
    printf("failed to suspend the thread after making the arbitrary call: %s\n", mach_error_string(err));
    return 0;
  }
  
  // restore the old state
  //#endif
  err = thread_set_state(thread_port, ARM_THREAD_STATE64, (thread_state_t)&saved_thread_state, saved_thread_stateCnt);
  if (err != KERN_SUCCESS){
    printf("error setting new thread state for hijacked thread: %s\n", mach_error_string(err));
    return 0;
  }
  
  usleep(100000);
  
  // resume the target thread so it continues running:
  err = thread_resume(thread_port);
  if (err != KERN_SUCCESS) {
    printf("error resuming hijacked thread after arbitrary call: %s\n", mach_error_string(err));
    return 0;
  }
  
  //uint64_t ret_val = fcall_thread_state.__x[0];
#endif
  
  
  // the thread is suspended
  //printf("remote function call return value: %llx\n", ret_val);
  
  return ret_val;
}

//#if 0


extern void _pthread_set_self(char*);
uint32_t task_remote_call(mach_port_t task_port, void* fptr, int n_params, ...)
{
  if (n_params > MAX_REMOTE_ARGS || n_params < 0){
    printf("unsupported number of arguments to remote function (%d)\n", n_params);
    return 0;
  }
  
  kern_return_t err;
  
  
  
  uint32_t remote_stack_base = 0;
  uint32_t remote_stack_size = 1*1024*1024;
  
  remote_stack_base = remote_alloc(task_port, remote_stack_size);
  
  uint32_t remote_stack_middle = remote_stack_base + (remote_stack_size/2);

  _STRUCT_ARM_THREAD_STATE thread_state = {0};
  mach_msg_type_number_t thread_stateCnt = sizeof(thread_state)/4;

  thread_state.__sp = remote_stack_middle;
  thread_state.__pc = (uint32_t)_pthread_set_self;
  thread_state.__cpsr = 0x40000010;
  
  uint32_t loop_pc = find_blx_r4_gadget();
  
  thread_state.__r[4] = loop_pc | 1;
  thread_state.__lr = loop_pc | 1;
  if (thread_state.__pc & 1) {
    thread_state.__cpsr |= 0x20;
  }
  
  thread_state.__r[0] = 0;
  
  mach_port_t thread_port = MACH_PORT_NULL;
  
  err = thread_create_running(task_port, ARM_THREAD_STATE, (thread_state_t)&thread_state, thread_stateCnt, &thread_port);
  if (err != KERN_SUCCESS){
    printf("error creating thread in target: %s\n", mach_error_string(err));
    return 0;
  }
  printf("new thread running in target: %x\n", thread_port);
  
  // wait for it to hit the loop:
  while(1){
    usleep(100*1000);
    // monitor the thread until we see it's in the infinite loop indicating it's done:
    err = thread_get_state(thread_port, ARM_THREAD_STATE, (thread_state_t)&thread_state, &thread_stateCnt);
    if (err != KERN_SUCCESS){
      printf("error getting thread state: %s\n", mach_error_string(err));
      return 0;
    }
    
    printf("thread_state.__pc: 0x%x\n", thread_state.__pc);
    
    if (thread_state.__pc == loop_pc){
      // thread has returned from the target function
      break;
    }
  }
  
  err = thread_suspend(thread_port);
  if (err != KERN_SUCCESS){
    printf("unable to suspend target thread\n");
    return 0;
  }
  
  // now make the arbitrary call
  
  memset(&thread_state, 0, sizeof(thread_state));
  
  thread_state.__sp = remote_stack_middle;
  thread_state.__pc = (uint32_t)fptr;
  thread_state.__cpsr = 0x40000010;
  
  thread_state.__r[4] = loop_pc | 1;
  thread_state.__lr = loop_pc | 1;
  if (thread_state.__pc & 1) {
    thread_state.__cpsr |= 0x20;
  }

  
  // load the arguments
  va_list ap;
  va_start(ap, n_params);
  
  arg_desc* args[MAX_REMOTE_ARGS] = {0};
  
  uint32_t remote_buffers[MAX_REMOTE_ARGS] = {0};
  
  for (int i = 0; i < n_params; i++){
    arg_desc* arg = va_arg(ap, arg_desc*);
    
    args[i] = arg;
    
    switch(arg->type){
      case ARG_LITERAL:
      {
        //printf("setting arg %d to literal %llx\n", i, arg->value);
        if (i < 4) {
          thread_state.__r[i] = arg->value;
        } else {
          uint32_t stack_slot = i-4;
          remote_write_32(task_port, thread_state.__sp+(stack_slot*4), arg->value);
        }
        break;
      }

      case ARG_BUFFER:
      case ARG_BUFFER_PERSISTENT:
      {
        uint32_t remote_buffer = alloc_and_fill_remote_buffer(task_port, arg->value, arg->length);
        if (i < 4) {
          thread_state.__r[i] = remote_buffer;
        } else {
          uint32_t stack_slot = i-4;
          remote_write_32(task_port, thread_state.__sp+(stack_slot*4), remote_buffer);
        }
        break;
      }

      case ARG_OUT_BUFFER:
      {
        uint32_t remote_buffer = remote_alloc(task_port, arg->length);
        printf("allocated a remote out buffer: 0x%x\n", remote_buffer);
        remote_buffers[i] = remote_buffer;
        if (i < 4) {
          thread_state.__r[i] = remote_buffer;
        } else {
          uint32_t stack_slot = i-4;
          remote_write_32(task_port, thread_state.__sp+(stack_slot*4), remote_buffer);
        }
        break;
      }

      default:
      {
        printf("invalid argument type!\n");
      }
    }
  }
  
  va_end(ap);
  //#if 0
  printf("fcall thread state:\n");
  printf("pc: 0x%08x\n", thread_state.__pc);
  printf("sp: 0x%08x\n", thread_state.__sp);
  printf("lr: 0x%08x\n", thread_state.__lr);
  printf("cpsr: 0x%08x\n", thread_state.__cpsr);
  for (int i = 0; i < 13; i++) {
    printf("x%d: 0x%08x\n", i, thread_state.__r[i]);
  }
  //#endif
  
  // set the thread state:
  err = thread_set_state(thread_port, ARM_THREAD_STATE, (thread_state_t)&thread_state, thread_stateCnt);
  if (err != KERN_SUCCESS){
    printf("error setting new thread state for hijacked thread: %s\n", mach_error_string(err));
    return 0;
  }
  
  // let the thread continue running with the new state:
  err = thread_resume(thread_port);
  if (err != KERN_SUCCESS) {
    printf("error resuming hijacked thread: %s\n", mach_error_string(err));
    return 0;
  }
  printf("resumed thread\n");
  
  // monitor for the function call ending and the thread hitting the infinite loop:
  // we're reusing fcall state so we can also get the return value via x0
  while(1){
    usleep(100*1000);
    thread_suspend(thread_port);
    err = thread_get_state(thread_port, ARM_THREAD_STATE, (thread_state_t)&thread_state, &thread_stateCnt);
    if (err != KERN_SUCCESS){
      printf("error getting thread state: %s\n", mach_error_string(err));
      return 0;
    }
    
    thread_resume(thread_port);
    
    if (thread_state.__pc == find_blx_r4_gadget()){
      // thread has returned from the target function
      printf("hit looper!\n");
      break;
    }
    printf("got bad pc: 0x%08x\n", thread_state.__pc);
  }
  
  
  
  // deallocate the remote thread
  err = thread_terminate(thread_port);
  if (err != KERN_SUCCESS){
    printf("failed to terminate thread\n");
    return 0;
  }
  mach_port_deallocate(mach_task_self(), thread_port);
  
  // handle post-call argument cleanup/copying:
  for (int i = 0; i < MAX_REMOTE_ARGS; i++){
    arg_desc* arg = args[i];
    if (arg == NULL){
      break;
    }
    switch (arg->type){
      case ARG_BUFFER:
      {
        remote_free(task_port, remote_buffers[i], arg->length);
        break;
      }

      case ARG_OUT_BUFFER:
      {
        // copy the contents back:
        remote_read_overwrite(task_port, remote_buffers[i], arg->value, arg->length);
        remote_free(task_port, remote_buffers[i], arg->length);
        break;
      }
    }
  }
  
  uint32_t ret_val = thread_state.__r[0];
  printf("ret val: %x\n", ret_val);
  
  // deallocate the stack in the target:
  remote_free(task_port, remote_stack_base, remote_stack_size);
  
  
  // thread is already suspended
  
  //thread_suspend(thread_port);
  
  //printf("suspended the thread...\n");
  //sleep(100);
  //thread_resume(thread_port);
#if 0
  // suspend the target thread:
  err = thread_suspend(thread_port);
  if (err != KERN_SUCCESS) {
    printf("failed to suspend the thread after making the arbitrary call: %s\n", mach_error_string(err));
    return 0;
  }
  
  // restore the old state
  //#endif
  err = thread_set_state(thread_port, ARM_THREAD_STATE64, (thread_state_t)&saved_thread_state, saved_thread_stateCnt);
  if (err != KERN_SUCCESS){
    printf("error setting new thread state for hijacked thread: %s\n", mach_error_string(err));
    return 0;
  }
  
  usleep(100000);
  
  // resume the target thread so it continues running:
  err = thread_resume(thread_port);
  if (err != KERN_SUCCESS) {
    printf("error resuming hijacked thread after arbitrary call: %s\n", mach_error_string(err));
    return 0;
  }
  
  //uint64_t ret_val = fcall_thread_state.__x[0];
#endif
  
  
  // the thread is suspended
  //printf("remote function call return value: %llx\n", ret_val);
  
  return ret_val;
}
//#endif
