#include "remote_memory.h"

#include <mach/mach.h>

extern kern_return_t vm_allocate
(
 vm_map_t target,
 vm_address_t *address,
 vm_size_t size,
 int flags
 );

extern kern_return_t vm_deallocate
(
 vm_map_t target,
 vm_address_t address,
 vm_size_t size
 );

extern kern_return_t vm_read_overwrite
(
 vm_map_t target_task,
 vm_address_t address,
 vm_size_t size,
 vm_address_t data,
 vm_size_t *outsize
 );

extern kern_return_t vm_write
(
 vm_map_t target_task,
 vm_address_t address,
 vm_offset_t data,
 mach_msg_type_number_t dataCnt
 );

uint32_t
remote_alloc(mach_port_t task_port,
             uint32_t size)
{
  kern_return_t err;
  
  vm_offset_t remote_addr = 0;
  vm_size_t remote_size = (vm_size_t)size;
  err = vm_allocate(task_port, &remote_addr, remote_size, 1); // ANYWHERE
  if (err != KERN_SUCCESS){
    printf("unable to allocate buffer in remote process\n");
    return 0;
  }
  return (uint32_t)remote_addr;
}

void
remote_free(mach_port_t task_port,
            uint32_t base,
            uint32_t size)
{
  kern_return_t err;
  
  err = vm_deallocate(task_port, (vm_address_t)base, (vm_size_t)size);
  if (err !=  KERN_SUCCESS){
    printf("unabble to deallocate remote buffer\n");
    return;
  }
  return;
}

void
remote_write(mach_port_t remote_task_port,
             uint32_t remote_address,
             uint32_t local_address,
             uint32_t length)
{
  kern_return_t err = vm_write(remote_task_port,
                                    (vm_address_t)remote_address,
                                    (vm_offset_t)local_address,
                                    (mach_msg_type_number_t)length);
  if (err != KERN_SUCCESS) {
    printf("remote write failed: %s %x\n", mach_error_string(err), err);
    return;
  }
}

void remote_write_32(mach_port_t task_port,
                     uint32_t remote_address,
                     uint32_t value)
{
  remote_write(task_port, remote_address, &value, 4);
}

uint32_t remote_read_32(mach_port_t task_port,
                        uint32_t remote_address)
{
  uint32_t val = 0;
  remote_read_overwrite(task_port, remote_address, &val, 4);
  return val;
}

uint32_t
alloc_and_fill_remote_buffer(mach_port_t task_port,
                             uint32_t local_address,
                             uint32_t length)
{
  kern_return_t err;
  
  uint32_t remote_address = remote_alloc(task_port, length);
  
  err = vm_write(task_port, remote_address, (vm_offset_t)local_address, (mach_msg_type_number_t)length);
  if (err != KERN_SUCCESS){
    printf("unable to write to remote memory\n");
    return 0;
  }
  
  return remote_address;
}

void
remote_read_overwrite(mach_port_t task_port,
                      uint32_t remote_address,
                      uint32_t local_address,
                      uint32_t length)
{
  kern_return_t err;
  
  vm_size_t outsize = 0;
  err = vm_read_overwrite(task_port, (vm_address_t)remote_address, (vm_size_t)length, (vm_address_t)local_address, &outsize);
  if (err != KERN_SUCCESS){
    printf("remote read failed\n");
    return;
  }
  
  if (outsize != length){
    printf("remote read was short (expected %x, got %x\n", length, outsize);
    return;
  }
}









